# -*- coding: utf-8 -*-
"""
Datebook month views
"""
import datetime

from django import http
from django.views import generic
from django.core.urlresolvers import reverse

from datebook.models import Datebook
from datebook.mixins import DatebookCalendarMixin
from datebook.forms.month import DatebookForm

class DatebookMonthAddView(DatebookCalendarMixin, generic.View):
    """
    Automatically create the datebook if it does not allready exists for the "author+year+month" 
    kwargs given, then redirect to its page.
    
    If the Datebook allready exists for the given kwargs, raise a "Http404"
    """
    def get(self, request, *args, **kwargs):
        if Datebook.objects.filter(author=self.author, period__year=self.year, period__month=self.month).count()>0:
            raise http.Http404
        
        d = self.author.datebook_set.create(period=datetime.date(year=self.year, month=self.month, day=1))
        
        return http.HttpResponseRedirect(d.get_absolute_url())


class DatebookMonthFormView(generic.FormView):
    """
    Datebook create form view
    """
    model = Datebook
    form_class = DatebookForm
    template_name = 'datebook/month/form.html'
    permission_required = 'datebook.add_datebook'
    raise_exception = True
    
    def form_valid(self, form):
        self.object = form.save()
        return super(DatebookMonthFormView, self).form_valid(form)
    
    def get_success_url(self):
        return self.object.get_absolute_url()



class DatebookMonthView(DatebookCalendarMixin, generic.TemplateView):
    """
    Datebook month details view
    
    Get the Calendar for the given year+month then fill it with day entries
    """
    template_name = "datebook/month/calendar.html"
    
    def get_calendar(self, day_filters={}):
        # Add current day if the datebook period is the current month+year
        current_day = None
        _curr = datetime.date.today()
        if _curr.replace(day=1) == self.object.period:
            current_day = _curr
        
        _cal = super(DatebookMonthView, self).get_calendar()
        
        return {
            "weekheader": _cal.formatweekheader(),
            "month": _cal.formatmonth(self.object.period.year, self.object.period.month, dayentries=self.get_dayentry_list(day_filters), current_day=current_day),
        }
        
    def get_context_data(self, **kwargs):
        context = super(DatebookMonthView, self).get_context_data(**kwargs)
        context.update({
            'datebook': self.object,
            'datebook_calendar': self.get_calendar(),
        })
        return context
    
    def get(self, request, *args, **kwargs):
        self.object = self.get_datebook({'period__year': self.year, 'period__month': self.month})
        
        context = self.get_context_data(**kwargs)
        
        return self.render_to_response(context)



# -*- coding: utf-8 -*-
"""
Calendar widgets
"""
import datetime
from calendar import TextCalendar, LocaleHTMLCalendar, _localized_day

class DatebookCalendar(TextCalendar):
    """
    Simple inherit from "TextCalendar" to fill calendars with datas from Datebook(s)
    
    Format methods generally return lists to be used within templates, not for 
    plain/text like TextCalendar
    """
    cssclasses = ["mon", "tue", "wed", "thu", "fri", "sat", "sun"]
    
    def formatweekheader(self):
        """
        Return a list of tuples ``(CLASSNAME, LOCALIZED_NAME)``
        """
        return [(self.cssclasses[i], _localized_day('%A')[i]) for i in self.iterweekdays()]

    def formatmonth(self, theyear, themonth, dayentries=None, current_day=None, withyear=True):
        """
        Return a list of weeks, each week is a list of day infos as dict
        """
        # Make a dict from the day entries, indexed on the date object
        entries_map = dict(map(lambda x: (x.activity_date, x), dayentries or []))
        # Walk in the weeks of the month to fill days with their associated DayEntry 
        # instance if any
        month = []
        for a, week in enumerate(self.monthdatescalendar(theyear, themonth)):
            month.append([])
            for i, day in enumerate(week):
                obj = None
                if day in entries_map: 
                    obj = entries_map.pop(day)
                month[a].append({
                    'date': day,
                    'entry': obj,
                    'noday': not(day.month==themonth),
                    'is_current': day==current_day,
                })
        return month



# -*- coding: utf-8 -*-
"""
Forms for year forms
"""
import datetime

from django import forms
from django.utils.translation import ugettext as _
from django.contrib.auth.models import User

from crispy_forms.helper import FormHelper
from crispy_forms_foundation.layout import Submit

from datebook.models import Datebook
from datebook.forms import CrispyFormMixin

class DatebookYearForm(CrispyFormMixin, forms.Form):
    """
    Datebook year form
    """
    crispy_form_helper_path = 'datebook.forms.crispies.year_helper'
    
    def __init__(self, author, queryset, *args, **kwargs):
        self.author = author
        
        super(DatebookYearForm, self).__init__(*args, **kwargs)
        super(forms.Form, self).__init__(*args, **kwargs)
        
        today = datetime.date.today()
        existing_years = [item.year for item in queryset]
        available_years = [(k, k) for k in range(today.year-3, today.year+4) if k not in existing_years]
        
        self.fields['year'] = forms.ChoiceField(label=_('year'), choices=available_years)
    
    def save(self, *args, **kwargs):
        instance = Datebook(
            author=self.author,
            period=datetime.date(int(self.cleaned_data['year']), 1, 1)
        )
        instance.save()
        
        return instance



# -*- coding: utf-8 -*-
"""
Forms for month forms
"""
from django import forms
from django.utils.translation import ugettext as _
from django.contrib.auth.models import User

from crispy_forms.helper import FormHelper
from crispy_forms_foundation.layout import Submit

from datebook.models import Datebook
from datebook.forms import CrispyFormMixin

class DatebookForm(CrispyFormMixin, forms.Form):
    """
    Datebook form
    """
    crispy_form_helper_path = 'datebook.forms.crispies.month_helper'
    
    def __init__(self, *args, **kwargs):
        super(DatebookForm, self).__init__(*args, **kwargs)
        super(forms.Form, self).__init__(*args, **kwargs)
        
        excluded_users = Datebook.objects.all().values('author_id').distinct()
        available_users = User.objects.all().exclude(pk__in=[v['author_id'] for v in excluded_users]).order_by('username')
        
        self.fields['owner'] = forms.ModelChoiceField(label=_('owner'), queryset=available_users, empty_label=None)
        self.fields['period'] = forms.DateField(label=_('period'))
    
    def clean_period(self):
        period = self.cleaned_data['period']
        period = period.replace(day=1)
        
        return period
    
    def save(self, *args, **kwargs):
        instance = Datebook(
            author=self.cleaned_data['owner'],
            period=self.cleaned_data['period']
        )
        instance.save()
        
        return instance



# -*- coding: utf-8 -*-
"""
Modèles de données
"""
import datetime

from django.db import models
from django.contrib.auth.models import User
from django.utils.translation import ugettext_lazy as _
from django.utils.timezone import now as tz_now, is_aware, localtime

from datebook import utils


class Datebook(models.Model):
    """
    Datebook is only for one user (its author) it is not shared. A Datebook is for a specific month of a year.
    
    A Datebook is alike "lazy" in the way that its 'DayEntry' are only created when the 
    user fill them, Datebook are not initialized with all its DayEntry on create.
    """
    author = models.ForeignKey(User, verbose_name=_('author'))
    created = models.DateTimeField(_('created'), blank=True, editable=False)
    modified = models.DateTimeField(_('last edit'), auto_now=True)
    period = models.DateField(_("month of activity"), blank=False)

    def __unicode__(self):
        return self.period.strftime("%m/%Y")

    @models.permalink
    def get_absolute_url(self):
        return ('datebook-author-month', (), {
            'author': self.author.username,
            'year': self.period.year,
            'month': self.period.strftime('%m'),
        })

    def clean(self):
        """Allways forcing to the first day of the month"""
        self.period = self.period.replace(day=1) 

    def save(self, *args, **kwargs):
        # First create
        if not self.created:
            self.created = tz_now()
        
        super(Datebook, self).save(*args, **kwargs)

    class Meta:
        unique_together = ("author", "period")
        verbose_name = _("Datebook")
        verbose_name_plural = _("Datebooks")


class DayEntry(models.Model):
    """
    Activity day in a Datebook
    """
    datebook = models.ForeignKey(Datebook, verbose_name=_('datebook'))
    activity_date = models.DateField(_('activity day date'), blank=False) # inherit month and year from its Datebook
    vacation = models.BooleanField(_('vacation'), default=False, blank=True, null=False)
    content = models.TextField(_("content"), max_length=500, blank=True) # for free ReST text

    start = models.DateTimeField(_('start'), blank=False) # should default to something like 9h with the dayentry date
    stop = models.DateTimeField(_('stop'), blank=False) # should default to something like 19h with the dayentry date
    pause = models.TimeField(_('pause'), blank=False, default=datetime.time(0, 0))

    def __unicode__(self):
        if not self.activity_date:
            return _("Empty")
        return self.activity_date.strftime("%d/%m/%Y")

    def get_display_hour(self, timeobj):
        """Formating time "XXhYY" display, minute are not displayed if not > 0"""
        _m = ""
        timeobj = localtime(timeobj)
        if timeobj.minute > 0:
            _m = str(timeobj.minute).rjust(2, "0")
        return _('%(hour)sh%(minute)s') % {'hour': str(timeobj.hour).rjust(2, "0"), 'minute': _m}

    def get_working_hours(self):
        """Display working hours like: '08h to 18h59'"""
        return _('%(start)s to %(stop)s') % {'start': self.get_display_hour(self.start), 'stop': self.get_display_hour(self.stop)}

    def get_elapsed_time(self):
        """Display elapsed time between start and stop, where pause time has been substracted from"""
        diff = utils.timedelta_to_seconds((self.stop-self.start))-utils.time_to_seconds(self.pause)
        return utils.format_seconds_to_clock(diff)

    def clean(self):
        from django.core.exceptions import ValidationError
        # Inherit the year and month from its datebook
        if hasattr(self, 'datebook'):
            self.activity_date = self.activity_date.replace(month=self.datebook.period.month, year=self.datebook.period.year)
        # DEPRECATED: Stuff below are for the form controller
        ## Stop can't be less than start
        #if self.stop < self.start:
            #raise ValidationError(_("Hour stop can't be less than hour start."))
        ## TODO: Pause time can't be superior to the time of (stop-start)

    def save(self, *args, **kwargs):
        # Allways update the datebook
        self.datebook.modified = tz_now()
        self.datebook.save()
        
        super(DayEntry, self).save(*args, **kwargs)

    class Meta:
        unique_together = ("datebook", "activity_date")
        verbose_name = _("day entry")
        verbose_name_plural = _("day entries")



# -*- coding: utf-8 -*-
"""
Some generic views
"""
from django.views.generic.list import ListView
from django.views.generic.edit import FormMixin

class ListAppendView(ListView, FormMixin):
    """
    A view to display an object list with a form to append a new object
    
    This view re-use some code from FormMixin and SimpleListView, sadly it seem not 
    possible to simply mix them.
    
    Need "model" and "form_class" attributes for the form parts and the required one 
    by BaseListView. "get_success_url" method should be filled too.
    
    "locked_form" is used to disable form (like if your list object is closed to new 
    object)
    """
    model = None
    form_class = None
    template_name = None
    paginate_by = None
    locked_form = False
    
    def form_valid(self, form):
        self.object = form.save()
        return super(ListAppendView, self).form_valid(form)

    def form_invalid(self, form):
        return self.render_to_response(self.get_context_data(object_list=self.object_list, form=form))

    def get_locked_form(self, form_class):
        return self.locked_form

    def get_form(self, form_class):
        """
        Returns an instance of the form to be used in this view.
        """
        if self.get_locked_form(form_class):
            return None
        return form_class(**self.get_form_kwargs())
        
    def get(self, request, *args, **kwargs):
        self.object_list = self.get_queryset()
        
        form_class = self.get_form_class()
        form = self.get_form(form_class)
        
        allow_empty = self.get_allow_empty()
        if not allow_empty and len(self.object_list) == 0:
            raise Http404(u"Empty list and '%(class_name)s.allow_empty' is False.".format(class_name=self.__class__.__name__))
        
        context = self.get_context_data(object_list=self.object_list, form=form)
        return self.render_to_response(context)

    def post(self, request, *args, **kwargs):
        self.object_list = self.get_queryset()
        
        form_class = self.get_form_class()
        form = self.get_form(form_class)
        
        allow_empty = self.get_allow_empty()
        if not allow_empty and len(self.object_list) == 0:
            raise Http404(u"Empty list and '%(class_name)s.allow_empty' is False.".format(class_name=self.__class__.__name__))
        
        if form and form.is_valid():
            return self.form_valid(form)
        elif form:
            return self.form_invalid(form)
        else:
            context = self.get_context_data(object_list=self.object_list, form=form)
            return self.render_to_response(context)

    # PUT is a valid HTTP verb for creating (with a known URL) or editing an
    # object, note that browsers only support POST for now.
    def put(self, *args, **kwargs):
        return self.post(*args, **kwargs)



# -*- coding: utf-8 -*-
"""
Day entry views
"""
import datetime

from django.conf import settings
from django.views import generic
from django.core.urlresolvers import reverse
from django.contrib.auth.models import User
from django.shortcuts import get_object_or_404

from braces.views import JSONResponseMixin

from datebook.models import Datebook, DayEntry
from datebook.mixins import DatebookCalendarMixin, DatebookCalendarAutoCreateMixin
from datebook.forms.day import DayEntryForm
from datebook.utils import week_from_date

class DayEntryBaseFormView(DatebookCalendarMixin):
    """
    DayEntry base form view
    """
    model = DayEntry
    context_object_name = "dayentry"
    template_name = "datebook/day/form.html"
    ajax_template_name = "datebook/day/form_fragment.html"
    form_class = DayEntryForm

    def get_form(self, form_class):
        """
        Add required args to form instance
        """
        return form_class(self.datebook, self.day, **self.get_form_kwargs())
    
    def switch_template(self, request):
        if request.is_ajax():
            return self.ajax_template_name
        return self.template_name
    
    def get(self, request, *args, **kwargs):
        self.template_name = self.switch_template(request)
        self.datebook = self.get_datebook({'period__year': self.year, 'period__month': self.month})
        
        return super(DayEntryBaseFormView, self).get(request, *args, **kwargs)
    
    def post(self, request, *args, **kwargs):
        self.template_name = self.switch_template(request)
        self.datebook = self.get_datebook({'period__year': self.year, 'period__month': self.month})
        
        return super(DayEntryBaseFormView, self).post(request, *args, **kwargs)
    
    def get_form_kwargs(self):
        """
        Add post form url
        """
        kwargs = super(DayEntryBaseFormView, self).get_form_kwargs()
        kwargs.update({'form_action': '.'})
        return kwargs

    def get_success_url(self):
        return reverse('datebook-author-month', kwargs={
            'author': self.author,
            'year': self.object.activity_date.year,
            'month': self.object.activity_date.month,
        })
        
    def get_next(self):
        return None


class DayEntryFormCreateView(DayEntryBaseFormView, generic.CreateView):
    """
    DayEntry form create view
    """
    def get_form_kwargs(self):
        """
        Put the initial data for the datetime start and stop
        
        NOTE: Time start and stop should comes from the User profile settings (date 
              allways come from the datebook period and the required day).
        """
        kwargs = super(DayEntryFormCreateView, self).get_form_kwargs()
        day_date = self.datebook.period.replace(day=self.day)
        time_start = datetime.time(9, 0)
        time_stop = datetime.time(18, 0)
        kwargs.update({
            'initial': {
                'start' : datetime.datetime.combine(day_date, time_start),
                'stop' : datetime.datetime.combine(day_date, time_stop),
                'pause' : datetime.time(1, 0),
            },
            'form_action': reverse('datebook-author-day-add', kwargs={
                'author': self.author,
                'year': self.year,
                'month': self.month,
                'day': self.day,
            })
        })
        return kwargs


class DayEntryFormEditView(DayEntryBaseFormView, generic.UpdateView):
    """
    DayEntry form edit view
    """
    def get_object(self):
        """
        Add required args to form instance
        """
        return self.datebook.dayentry_set.get(activity_date__day=self.day)
    
    def get_form_kwargs(self):
        """
        Add post form url
        """
        kwargs = super(DayEntryBaseFormView, self).get_form_kwargs()
        kwargs.update({
            'form_action': reverse('datebook-author-day-edit', kwargs={
                'author': self.author,
                'year': self.year,
                'month': self.month,
                'day': self.day,
            })
        })
        return kwargs


class DayEntryCurrentView(generic.RedirectView):
    permanent = False
    
    def get_current_date(self):
        """
        Get the current date
        """
        today = datetime.date.today()
        self.year = self.kwargs['year'] = today.year
        self.month = self.kwargs['month'] = today.month
        self.day = self.kwargs['day'] = today.day
        return today
    
    def get_datebook(self):
        """
        Get or Create datebook
        """
        try:
            obj = Datebook.objects.get(author__username=self.kwargs['author'], period__year=self.year, period__month=self.month)
        except Datebook.DoesNotExist:
            self.author = get_object_or_404(User, username=self.kwargs['author'])
            obj = Datebook(author=self.author, period=datetime.date(self.year, self.month, 1))
            obj.save()
        else:
            self.author = obj.author
        return obj
    
    def get_redirect_url(self, author):
        """
        Get Create or Edit dayentry url
        """
        today = self.get_current_date()
        datebook = self.get_datebook()
        
        try:
            obj = DayEntry.objects.get(datebook=datebook, activity_date=datetime.date(self.year, self.month, self.day))
        except DayEntry.DoesNotExist:
            return reverse('datebook-author-day-add', kwargs={
                'author': self.author,
                'year': self.year,
                'month': self.month,
                'day': self.day,
            })
        else:
            return reverse('datebook-author-day-edit', kwargs={
                'author': self.author,
                'year': self.year,
                'month': self.month,
                'day': self.day,
            })


class DayEntryDetailView(DatebookCalendarMixin, generic.TemplateView):
    """
    DayEntry detail view
    """
    model = DayEntry
    template_name = "datebook/day/detail_fragment.html"
        
    def get_previous(self):
        try:
            obj = self.object.get_previous_by_activity_date(**{
                'activity_date__month':self.object.activity_date.month,
                'activity_date__year':self.object.activity_date.year,
            })
        except DayEntry.DoesNotExist:
            pass
        else:
            return obj
        return None
        
    def get_next(self):
        try:
            obj = self.object.get_next_by_activity_date(**{
                'activity_date__month':self.object.activity_date.month,
                'activity_date__year':self.object.activity_date.year,
            })
        except DayEntry.DoesNotExist:
            pass
        else:
            return obj
        return None
        
    def get_object(self, **kwargs):
        try:
            obj = DayEntry.objects.get(datebook=self.datebook, activity_date=datetime.date(self.year, self.month, self.day))
        except DayEntry.DoesNotExist:
            pass
        else:
            return obj
        return None
    
    def get_context_data(self, **kwargs):
        context = super(DayEntryDetailView, self).get_context_data(**kwargs)
        context.update({
            'datebook': self.datebook,
            'object': self.object,
            'previous_day': self.previous_day,
            'next_day': self.next_day,
        })
        return context
    
    def get(self, request, *args, **kwargs):
        self.datebook = self.get_datebook({'period__year': self.year, 'period__month': self.month})
        self.object = self.get_object()
        self.previous_day = self.get_previous()
        self.next_day = self.get_next()
        
        context = self.get_context_data(**kwargs)
        
        return self.render_to_response(context)




# -*- coding: utf-8 -*-
"""
Application Crumbs
"""
from autobreadcrumbs import site
from django.utils.translation import ugettext_lazy

site.update({
    'datebook-index': ugettext_lazy('Datebooks'),
    'datebook-create': ugettext_lazy('Create a datebook'),
    'datebook-author': '{{ author.username }}',
    'datebook-author-year': '{{ year }}',
    'datebook-author-month': '{{ target_date|date:"F" }}',
    'datebook-author-day-add': ugettext_lazy('Add day {{ day }}'),
    'datebook-author-day-edit': ugettext_lazy('Edit day {{ day }}'),
})



# -*- coding: utf-8 -*-
"""
Datebook year views
"""
import datetime
import calendar

from django.views import generic

from datebook.mixins import DateKwargsMixin

class DatebookYearView(DateKwargsMixin, generic.TemplateView):
    """
    Datebook year view
    
    Display the twelve months of the given year with link and infos for the 
    existing datebooks
    """
    template_name = "datebook/year.html"
        
    def get_context_data(self, **kwargs):
        context = super(DatebookYearView, self).get_context_data(**kwargs)
        
        _curr = datetime.date.today()
        # Get all datebooks for the given year
        queryset = self.object.datebook_set.filter(period__year=self.year).order_by('period')[0:13]
        _datebook_map = dict(map(lambda x: (x.period.month, x), queryset))
        # Fill the finded datebooks in the month map, month without datebook will have 
        # None instead of a Datebook instance
        datebooks_map = [(datetime.datetime(self.year, i, 1), _datebook_map.get(i)) for i in range(1,13)]
        
        context.update({
            'year_current': _curr.year,
            'is_current_year': (self.year == _curr.year),
            'datebooks_map': datebooks_map,
        })
        return context
    
    def get(self, request, *args, **kwargs):
        self.object = self.author
        
        context = self.get_context_data(**kwargs)
        
        return self.render_to_response(context)

